from . import STDR

class EPS(STDR):
    def __init__(self, version=None, endian=None, record=None):
        self.id = 'EPS'
        self.local_debug = False
        # Version
        if version==None or version=='V4' or version=='V3':
            self.version = version
            self.info = '''
End Program Section Record
--------------------------

Function:
    Marks the end of the current program section (or sequencer) in the job plan.

Frequency:
    * Optional on each exit from the program segment.

Location:
    Following the corresponding BPS and before the PRR in the data stream.
'''
#           Changed by seimit : to pass serialization test, set None value for 'Ref' instead the old ''  
            self.fields = {
                'REC_LEN'  : {'#' :  0, 'Type' :  'U*2', 'Ref' :         None, 'Value' :   None, 'Text' : 'Bytes of data following header        ', 'Missing' : 'N/A                    '},
                'REC_TYP'  : {'#' :  1, 'Type' :  'U*1', 'Ref' :         None, 'Value' :     20, 'Text' : 'Record type                           ', 'Missing' : '20                     '},
                'REC_SUB'  : {'#' :  2, 'Type' :  'U*1', 'Ref' :         None, 'Value' :     20, 'Text' : 'Record sub-type                       ', 'Missing' : '10                     '}
            }
        else:
            raise STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
        self._default_init(endian, record)



import sys
from . import STDR

class FTR(STDR):
    def __init__(self, version=None, endian=None, record=None):
        self.id = 'FTR'
        self.local_debug = False
        if version==None or version=='V4':
            self.version = 'V4'
            self.info = '''
Functional Test Record
----------------------

Function:
    Contains the results of the single execution of a functional test in the test program. The
    first occurrence of this record also establishes the default values for all semi-static
    information about the test. The FTR is related to the Test Synopsis Record (TSR) by test
    number, head number, and site number.

Frequency:
    * Obligatory, one or more for each execution of a functional test.

Location:
    Anywhere in the data stream after the corresponding Part Information Record (PIR)
    and before the corresponding Part Result Record (PRR).
'''
            self.fields = {
                'REC_LEN'  : {'#' :  0, 'Type' : 'U*2',  'Ref' :       None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' :    None},
                'REC_TYP'  : {'#' :  1, 'Type' : 'U*1',  'Ref' :       None, 'Value' :   15, 'Text' : 'Record type                           ', 'Missing' :    None},
                'REC_SUB'  : {'#' :  2, 'Type' : 'U*1',  'Ref' :       None, 'Value' :   20, 'Text' : 'Record sub-type                       ', 'Missing' :    None},
                'TEST_NUM' : {'#' :  3, 'Type' : 'U*4',  'Ref' :       None, 'Value' : None, 'Text' : 'Test number                           ', 'Missing' :    None}, # Obligatory!
                'HEAD_NUM' : {'#' :  4, 'Type' : 'U*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Test head number                      ', 'Missing' :       1},
                'SITE_NUM' : {'#' :  5, 'Type' : 'U*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Test site number                      ', 'Missing' :       1},
                'TEST_FLG' : {'#' :  6, 'Type' : 'B*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Test flags (fail, alarm, etc.)        ', 'Missing' : ['0']*8},
                'OPT_FLAG' : {'#' :  7, 'Type' : 'B*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Optional data flag                    ', 'Missing' : ['1']*8},
                'CYCL_CNT' : {'#' :  8, 'Type' : 'U*4',  'Ref' :       None, 'Value' : None, 'Text' : 'Cycle count of vector                 ', 'Missing' :       0}, # OPT_FLAG bit0 = 1
                'REL_VADR' : {'#' :  9, 'Type' : 'U*4',  'Ref' :       None, 'Value' : None, 'Text' : 'Relative vector address               ', 'Missing' :       0}, # OPT_FLAG bit1 = 1
                'REPT_CNT' : {'#' : 10, 'Type' : 'U*4',  'Ref' :       None, 'Value' : None, 'Text' : 'Repeat count of vector                ', 'Missing' :       0}, # OPT_FLAG bit2 = 1
                'NUM_FAIL' : {'#' : 11, 'Type' : 'U*4',  'Ref' :       None, 'Value' : None, 'Text' : 'Number of pins with 1 or more failures', 'Missing' :       0}, # OPT_FLAG bit3 = 1
                'XFAIL_AD' : {'#' : 12, 'Type' : 'I*4',  'Ref' :       None, 'Value' : None, 'Text' : 'X logical device failure address      ', 'Missing' :       0}, # OPT_FLAG bit4 = 1
                'YFAIL_AD' : {'#' : 13, 'Type' : 'I*4',  'Ref' :       None, 'Value' : None, 'Text' : 'Y logical device failure address      ', 'Missing' :       0}, # OPT_FLAG bit4 = 1
                'VECT_OFF' : {'#' : 14, 'Type' : 'I*2',  'Ref' :       None, 'Value' : None, 'Text' : 'Offset from vector of interest        ', 'Missing' :       0}, # OPT_FLAG bit5 = 1
                'RTN_ICNT' : {'#' : 15, 'Type' : 'U*2',  'Ref' :       None, 'Value' : None, 'Text' : 'Count (j) of return data PMR indexes  ', 'Missing' :       0},
                'PGM_ICNT' : {'#' : 16, 'Type' : 'U*2',  'Ref' :       None, 'Value' : None, 'Text' : 'Count (k) of programmed state indexes ', 'Missing' :       0},
                'RTN_INDX' : {'#' : 17, 'Type' : 'xU*2', 'Ref' : 'RTN_ICNT', 'Value' : None, 'Text' : 'Array of j return data PMR indexes    ', 'Missing' :      []}, # RTN_ICNT = 0
                'RTN_STAT' : {'#' : 18, 'Type' : 'xN*1', 'Ref' : 'RTN_ICNT', 'Value' : None, 'Text' : 'Array of j returned states            ', 'Missing' :      []}, # RTN_ICNT = 0
                'PGM_INDX' : {'#' : 19, 'Type' : 'xU*2', 'Ref' : 'PGM_ICNT', 'Value' : None, 'Text' : 'Array of k programmed state indexes   ', 'Missing' :      []}, # PGM_ICNT = 0
                'PGM_STAT' : {'#' : 20, 'Type' : 'xN*1', 'Ref' : 'PGM_ICNT', 'Value' : None, 'Text' : 'Array of k programmed states          ', 'Missing' :      []}, # PGM_ICNT = 0
                'FAIL_PIN' : {'#' : 21, 'Type' : 'D*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Failing pin bitfield                  ', 'Missing' :      []},
                'VECT_NAM' : {'#' : 22, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Vector module pattern name            ', 'Missing' :      ''},
                'TIME_SET' : {'#' : 23, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Time set name                         ', 'Missing' :      ''},
                'OP_CODE'  : {'#' : 24, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Vector Op Code                        ', 'Missing' :      ''},
                'TEST_TXT' : {'#' : 25, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Descriptive text or label             ', 'Missing' :      ''},
                'ALARM_ID' : {'#' : 26, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Name of alarm                         ', 'Missing' :      ''},
                'PROG_TXT' : {'#' : 27, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Additional programmed information     ', 'Missing' :      ''},
                'RSLT_TXT' : {'#' : 28, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Additional result information         ', 'Missing' :      ''},
                'PATG_NUM' : {'#' : 29, 'Type' : 'U*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Pattern generator number              ', 'Missing' :    0xFF},
                'SPIN_MAP' : {'#' : 30, 'Type' : 'D*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Bit map of enabled comparators        ', 'Missing' :      []}
            }

        else:
            raise STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
        self._default_init(endian, record)

    def to_atdf(self):
        '''
        Method that writes A(SCII)TDF version of the STDF file.
        '''
        sequence = {}
        header = ''
        body = ''
        
        header = self.id + ':'
        
        #    The order of fields is different in STDF and ATDF for FTR record
            
        #    STDF page 57| ATDF page 51
            
        #     3 TEST_NUM    =  3 TEST_NUM   
        #     4 HEAD_NUM    =  4 HEAD_NUM    
        #     5 SITE_NUM    =  5 SITE_NUM  
        #     6 TEST_FLG    -> 6 TEST_FLG bits 6 & 7
        #                   -> 6 TEST_FLG bits 0, 2, 3, 4, & 5
        #     7 OPT_FLAG    -> missing    
        #     8 CYCL_CNT
        #     9 REL_VADR
        #    10 REPT_CNT
        #    11 NUM_FAIL
        #    12 XFAIL_AD
        #    13 YFAIL_AD  
        #    14 VECT_OFF
        #    15 RTN_ICNT    -> missing 
        #    16 PGM_ICNT    -> missing 
        #    17 RTN_INDX 
        #    18 RTN_STAT
        #    19 PGM_INDX
        #    20 PGM_STAT
        #    21 FAIL_PIN
        #    22 VECT_NAM    = 22 VECT_NAM
        #    23 TIME_SET    = 23 TIME_SET
        #    24 OP_CODE
        #    25 TEST_TXT
        #    26 ALARM_ID
        #    27 PROG_TXT
        #    28 RSLT_TXT
        #    29 PATG_NUM
        #    30 SPIN_MAP
        #                   =  8 CYCL_CNT
        #                   =  9 REL_VADR
        #                   = 10 REPT_CNT
        #                   = 11 NUM_FAIL
        #                   = 12 XFAIL_AD
        #                   = 13 YFAIL_AD
        #                   = 14 VECT_OFF
        #                   = 17 RTN_INDX
        #                   = 18 RTN_STAT
        #                   = 19 PGM_INDX
        #                   = 20 PGM_STAT
        #                   = 21 FAIL_PIN
        #                   = 24 OP_CODE
        #                   = 25 TEST_TXT
        #                   = 26 ALARM_ID
        #                   = 27 PROG_TXT
        #                   = 28 RSLT_TXT
        #                   = 29 PATG_NUM
        #                   = 30 SPIN_MAP

#       3 TEST_NUM 
        body += self.gen_atdf(3)
#       4 HEAD_NUM 
        body += self.gen_atdf(4)
#       5 SITE_NUM  
        body += self.gen_atdf(5)
        
#       6 TEST_FLG bits 6 & 7
#           bit 6: Pass/fail flag (bit 7) is valid
        v = self.get_fields(6)[3]
        if v != None and v[6] == '0':
#           bit 7:
#           0 = Part passed
            if self.get_fields(6)[3][7] == '0':
                body += 'P|'
#           1 = Part failed
            elif self.get_fields(6)[3][7] == '1':
                body += 'F|'
#                             6 TEST_FLG bits 0, 2, 3, 4, & 5
#           bit 0:
#           0 = No alarm
            if self.get_fields(6)[3][0] == '0':
                body += ''
#           1 = Alarm detected during testing
            elif self.get_fields(6)[3][0] == '1':
                body += 'A'
#           bit 2:
#           0= Test result is reliable
            if self.get_fields(6)[3][2] == '0':
                body += ''
#           1 = Test result is unreliable
            elif self.get_fields(6)[3][2] == '1':
                body += 'U'
#           bit 3:
#           0 = No timeout
            if self.get_fields(6)[3][3] == '0':
                body += ''
#           1 = Timeout occurred
            elif self.get_fields(6)[3][3] == '1':
                body += 'T'
#           bit 4:
#           0 = Test was executed
            if self.get_fields(6)[3][4] == '0':
                body += ''
#           1= Testnotexecuted
            elif self.get_fields(6)[3][4] == '1':
                body += 'N'
#           bit 5:
#           0 = No abort
            if self.get_fields(6)[3][5] == '0':
                body += '|'
#           1= Testaborted
            elif self.get_fields(6)[3][5] == '1':
                body += 'X|'
        else:
            body += '|'
            
#       22 VECT_NAM        
        body += self.gen_atdf(22)
#       23 TIME_SET
        body += self.gen_atdf(23)
#        8 CYCL_CNT
        body += self.gen_atdf(8)
#        9 REL_VADR
        body += self.gen_atdf(9)
#       10 REPT_CNT
        body += self.gen_atdf(10)
#       11 NUM_FAIL
        body += self.gen_atdf(11)
#       12 XFAIL_AD
        body += self.gen_atdf(12)
#       13 YFAIL_AD
        body += self.gen_atdf(13)
#       14 VECT_OFF
        body += self.gen_atdf(14)
#       17 RTN_INDX
        body += self.gen_atdf(17)
#       18 RTN_STAT
        body += self.gen_atdf(18)
#       19 PGM_INDX
        body += self.gen_atdf(19)
#       20 PGM_STAT
        body += self.gen_atdf(20)
#       22 FAIL_PIN
        body += self.gen_atdf(21)
#       24 OP_CODE
        body += self.gen_atdf(24)
#       25 TEST_TXT
        body += self.gen_atdf(25)
#       26 ALARM_ID
        body += self.gen_atdf(26)
#       27 PROG_TXT
        body += self.gen_atdf(27)
#       28 RSLT_TXT
        body += self.gen_atdf(28)
#       29 PATG_NUM
        body += self.gen_atdf(29)
#       30 SPIN_MAP
        body += self.gen_atdf(30)
        body = body[:-1] 

        # assemble the record
        retval = header + body

        if self.local_debug: print("%s._to_atdf()\n   '%s'\n" % (self.id, retval))
        return retval


from . import STDR

class ATR(STDR):
    def __init__(self, version=None, endian=None, record = None):
        self.id = 'ATR'
        self.local_debug = False
        # Version
        if version==None or version=='V4':
            self.version = 'V4'
            self.info=    '''
Audit Trail Record
------------------

Function:
    Used to record any operation that alters the contents of the STDF file. The name of the
    program and all its parameters should be recorded in the ASCII field provided in this
    record. Typically, this record will be used to track filter programs that have been
    applied to the data.

Frequency:
    * Optional
    * One for each filter or other data transformation program applied to the STDF data.

Location:
    Between the File Attributes Record (FAR) and the Master Information Record (MIR).
    The filter program that writes the altered STDF file must write its ATR immediately
    after the FAR (and hence before any other ATRs that may be in the file). In this way,
    multiple ATRs will be in reverse chronological order.
'''
            self.fields = {
                'REC_LEN'  : {'#' :  0, 'Type' : 'U*2' , 'Ref' : None, 'Value' :   None, 'Text' : 'Bytes of data following header        ', 'Missing' : None, 'Note' : ''},
                'REC_TYP'  : {'#' :  1, 'Type' : 'U*1' , 'Ref' : None, 'Value' :      0, 'Text' : 'Record type                           ', 'Missing' : None, 'Note' : ''},
                'REC_SUB'  : {'#' :  2, 'Type' : 'U*1' , 'Ref' : None, 'Value' :     20, 'Text' : 'Record sub-type                       ', 'Missing' : None, 'Note' : ''},
                'MOD_TIM'  : {'#' :  3, 'Type' : 'U*4' , 'Ref' : None, 'Value' :   None, 'Text' : 'Date & time of STDF file modification ', 'Missing' :    0, 'Note' : ''},
                'CMD_LINE' : {'#' :  4, 'Type' : 'C*n' , 'Ref' : None, 'Value' :   None, 'Text' : 'Command line of program               ', 'Missing' :   '', 'Note' : ''}
            }
        else:
            raise STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
        self._default_init(endian, record)


from . import STDR

class BPS(STDR):
    def __init__(self, version=None, endian=None, record=None):
        self. id = 'BPS'
        self.local_debug = False
        # Version
        if version==None or version=='V4' or version=='V3':
            if version==None: self.verson = 'V4'
            else: self.version = version
            self.info = '''
Begin Program Section Record
----------------------------

Function:
    Marks the beginning of a new program section (or sequencer) in the job plan.

Frequency:
    * Optional on each entry into the program segment.

Location:
    Anywhere after the PIR and before the PRR.
'''
            self.fields = {
                'REC_LEN'  : {'#' :  0, 'Type' :  'U*2', 'Ref' : None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' : None, 'Note' : ''},
                'REC_TYP'  : {'#' :  1, 'Type' :  'U*1', 'Ref' : None, 'Value' :   20, 'Text' : 'Record type                           ', 'Missing' : None, 'Note' : ''},
                'REC_SUB'  : {'#' :  2, 'Type' :  'U*1', 'Ref' : None, 'Value' :   10, 'Text' : 'Record sub-type                       ', 'Missing' : None, 'Note' : ''},
                'SEQ_NAME' : {'#' :  3, 'Type' :  'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Sequence name                         ', 'Missing' :   '', 'Note' : ''}
            }
        else:
            raise STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
        self._default_init(endian, record)



import sys
from . import STDR

class GDR(STDR):
    def __init__(self, version=None, endian=None, record=None):
        self.id = 'GDR'
        self.local_debug = False
        if version==None or version=='V4' or version=='V3':
            if version==None: self.version='V4'
            else: self.version=version
            self.info = '''
Generic Data Record
-------------------

Function:
    Contains information that does not conform to any other record type defined by the
    STDF specification. Such records are intended to be written under the control of job
    plans executing on the tester. This data may be used for any purpose that the user
    desires.

Frequency:
    * Optional, a test data file may contain any number of GDRs.

Location:
    Anywhere in the data stream after the initial "FAR-(ATRs)-MIR-(RDR)-(SDRs)" sequence.
'''
            self.fields = {
                'REC_LEN'  : {'#' : 0, 'Type' :  'U*2', 'Ref' :      None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' : None},
                'REC_TYP'  : {'#' : 1, 'Type' :  'U*1', 'Ref' :      None, 'Value' :   50, 'Text' : 'Record type                           ', 'Missing' : None},
                'REC_SUB'  : {'#' : 2, 'Type' :  'U*1', 'Ref' :      None, 'Value' :   10, 'Text' : 'Record sub-type                       ', 'Missing' : None},
                'FLD_CNT'  : {'#' : 3, 'Type' :  'U*2', 'Ref' :      None, 'Value' : None, 'Text' : 'Count of data fields in record        ', 'Missing' :    0},
                'GEN_DATA' : {'#' : 4, 'Type' : 'xV*n', 'Ref' : 'FLD_CNT', 'Value' : None, 'Text' : 'Data type code and data for one field ', 'Missing' :   []}
            }
        else:
            raise STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
        self._default_init(endian, record)

    def to_atdf(self):

        sequence = {}
        header = ''
        body = ''
        
        header = self.id + ':'

        atdf_code = ['', 'U', 'M', 'B', 'I', 'S', 'L', 'F', 'D', '', 'T', 'X', 'Y', 'N']
        pair = self.get_fields(4)[3]
        for i in range(len(pair)):
            code = pair[i][0][0]
            value = pair[i][0][1]
            body += atdf_code[code]
            if code == 11:
                body += self.hexify(value).replace('0x', '')
            elif code == 12:
                # HEX representation of bits
                bits_to_pack = len(value)
                bytes_to_pack = int(bits_to_pack) // 8
                if (bits_to_pack % 8) != 0:
                    bytes_to_pack += 1
                for Byte in range(bytes_to_pack):
                    byte = 0
                    for Bit in range(8):
                        if (Byte * 8) + Bit < len(value):
                            if value[(Byte * 8) + Bit] == '1':
                                byte += pow(2, Bit)
                    body += '%02X' % byte
            elif code == 13:
                # One nubble
                # ATDF is not clear if we can write only one nubble or two
                # so here we will output 2 nubbles seprated by comma
                for i in range(len(value)):
                    body += str(value[i]) + ','
                body = body[:-1] 
                
            else:
                body += str(value)
            body += "|"
        body = body[:-1] 

        # assemble the record
        retval = header + body

        if self.local_debug: print("%s._to_atdf()\n   '%s'\n" % (self.id, retval))
        return retval


import sys
import os
import time

from . import STDR

class WRR(STDR):
    def __init__(self, version=None, endian=None, record=None):
        self.id = 'WRR'
        self.local_debug = False
        if version==None or version=='V4':
            self.version = 'V4'
            self.info = '''
Wafer Results Record
--------------------

Function:
    Contains the result information relating to each wafer tested by the job plan. The WRR
    and the Wafer Information Record (WIR) bracket all the stored information pertaining
    to one tested wafer. This record is used only when testing at wafer probe time. A
    WIR/WRR pair will have the same HEAD_NUM and SITE_GRP values.

Frequency:
    * Obligatory for Wafer sort
    * One per wafer tested.

Location:
    Anywhere in the data stream after the corresponding WIR.
    Sent after testing each wafer.
'''
            self.fields = {
                'REC_LEN'  : {'#' :  0, 'Type' : 'U*2', 'Ref' : None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' : None      },
                'REC_TYP'  : {'#' :  1, 'Type' : 'U*1', 'Ref' : None, 'Value' :    2, 'Text' : 'Record type                           ', 'Missing' : None      },
                'REC_SUB'  : {'#' :  2, 'Type' : 'U*1', 'Ref' : None, 'Value' :   20, 'Text' : 'Record sub-type                       ', 'Missing' : None      },
                'HEAD_NUM' : {'#' :  3, 'Type' : 'U*1', 'Ref' : None, 'Value' : None, 'Text' : 'Test head number                      ', 'Missing' : 255       },
                'SITE_GRP' : {'#' :  4, 'Type' : 'U*1', 'Ref' : None, 'Value' : None, 'Text' : 'Site group number                     ', 'Missing' : 255       },
                'FINISH_T' : {'#' :  5, 'Type' : 'U*4', 'Ref' : None, 'Value' : None, 'Text' : 'Date and time last part tested        ', 'Missing' : 0         },
                'PART_CNT' : {'#' :  6, 'Type' : 'U*4', 'Ref' : None, 'Value' : None, 'Text' : 'Number of parts tested                ', 'Missing' : 0         },
                'RTST_CNT' : {'#' :  7, 'Type' : 'U*4', 'Ref' : None, 'Value' : None, 'Text' : 'Number of parts retested              ', 'Missing' : 4294967295},
                'ABRT_CNT' : {'#' :  8, 'Type' : 'U*4', 'Ref' : None, 'Value' : None, 'Text' : 'Number of aborts during testing       ', 'Missing' : 4294967295},
                'GOOD_CNT' : {'#' :  9, 'Type' : 'U*4', 'Ref' : None, 'Value' : None, 'Text' : 'Number of good (passed) parts tested  ', 'Missing' : 4294967295},
                'FUNC_CNT' : {'#' : 10, 'Type' : 'U*4', 'Ref' : None, 'Value' : None, 'Text' : 'Number of functional parts tested     ', 'Missing' : 4294967295},
                'WAFER_ID' : {'#' : 11, 'Type' : 'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Wafer ID                              ', 'Missing' : ''        },
                'FABWF_ID' : {'#' : 12, 'Type' : 'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Fab wafer ID                          ', 'Missing' : ''        },
                'FRAME_ID' : {'#' : 13, 'Type' : 'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Wafer frame ID                        ', 'Missing' : ''        },
                'MASK_ID'  : {'#' : 14, 'Type' : 'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Wafer mask ID                         ', 'Missing' : ''        },
                'USR_DESC' : {'#' : 15, 'Type' : 'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Wafer description supplied by user    ', 'Missing' : ''        },
                'EXC_DESC' : {'#' : 16, 'Type' : 'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Wafer description supplied by exec    ', 'Missing' : ''        }
            }

        else:
            raise STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
        self._default_init(endian, record)

    def to_atdf(self):

        sequence = {}
        header = ''
        body = ''
        
        header = self.id + ':'

#        The order of fields is different in STDF and ATDF for WRR record
#        
#        STDF page 38| ATDF page 34 
#        
#        3  HEAD_NUM    =  3  HEAD_NUM
#        4  SITE_GRP    
#        5  FINISH_T    =  5  FINISH_T
#        6  PART_CNT    =  6  PART_CNT   
#                        |11  WAFER_ID
#                        | 4  SITE_GRP
#        7  RTST_CNT    =  7  RTST_CNT
#        8  ABRT_CNT    =  8  ABRT_CNT
#        9  GOOD_CNT    =  9  GOOD_CNT
#        10 FUNC_CNT    = 10  FUNC_CNT
#        11 WAFER_ID    
#        12 FABWF_ID    = 12  FABWF_ID
#        13 FRAME_ID    = 13  FRAME_ID
#        14 MASK_ID     = 14  MASK_ID
#        15 USR_DESC    = 15  USR_DESC
#        16 EXC_DESC    = 16  EXC_DESC
    
#       3 HEAD_NUM
        body += self.gen_atdf(3)

#       5 FINISH_T
        v = self.get_fields(5)[3]
        if v != None:
            t = ""
            if os.name == "nt":
                t = time.strftime("%#H:%#M:%#S %#d-%b-%Y", time.gmtime(v))
            else:
                t = time.strftime("%-H:%-M:%-S %-d-%b-%Y", time.gmtime(v))
            body += "%s|" % (t.upper())

#       6 PART_CNT
        body += self.gen_atdf(6)
        
#       11 PART_CNT
        body += self.gen_atdf(11)

#       4 SITE_GRP
        body += self.gen_atdf(4)

#       7 RTST_CNT
        body += self.gen_atdf(7)

#       8 ABRT_CNT
        body += self.gen_atdf(8)

#       9 ABRT_CNT
        body += self.gen_atdf(9)

#       10 FUNC_CNT
        body += self.gen_atdf(10)

#       12 FABWF_ID
        body += self.gen_atdf(12)

#       13 FRAME_ID
        body += self.gen_atdf(13)

#       14 MASK_ID
        body += self.gen_atdf(14)

#       15 USR_DESC
        body += self.gen_atdf(15)

#       16 EXC_DESC
        body += self.gen_atdf(16)

        body = body[:-1]

        # assemble the record
        retval = header + body

        if self.local_debug: print("%s._to_atdf()\n   '%s'\n" % (self.id, retval))
        return retval



import sys
from . import STDR

class WCR(STDR):
    def __init__(self, version=None, endian=None, record=None):
        self.id = 'WCR'
        self.local_debug = False
        if version==None or version=='V4':
            if version==None: self.version='V4'
            else: self.version = version
            self.info = '''
Wafer Configuration Record
--------------------------

Function:
    Contains the configuration information for the wafers tested by the job plan. The
    WCR provides the dimensions and orientation information for all wafers and dice
    in the lot. This record is used only when testing at wafer probe time.

Frequency:
    * Obligatory for Wafer sort
    * One per STDF file

Location:
    Anywhere in the data stream after the initial "FAR-(ATRs)-MIR-(RDR)-(SDRs)" sequence, and before the MRR.
'''
            self.fields = {
                'REC_LEN'  : {'#' :  0, 'Type' : 'U*2', 'Ref' : None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' : None  },
                'REC_TYP'  : {'#' :  1, 'Type' : 'U*1', 'Ref' : None, 'Value' :    2, 'Text' : 'Record type                           ', 'Missing' : None  },
                'REC_SUB'  : {'#' :  2, 'Type' : 'U*1', 'Ref' : None, 'Value' :   30, 'Text' : 'Record sub-type                       ', 'Missing' : None  },
                'WAFR_SIZ' : {'#' :  3, 'Type' : 'R*4', 'Ref' : None, 'Value' : None, 'Text' : 'Diameter of wafer in WF_UNITS         ', 'Missing' : 0.0   },
                'DIE_HT'   : {'#' :  4, 'Type' : 'R*4', 'Ref' : None, 'Value' : None, 'Text' : 'Height of die in WF_UNITS             ', 'Missing' : 0.0   },
                'DIE_WID'  : {'#' :  5, 'Type' : 'R*4', 'Ref' : None, 'Value' : None, 'Text' : 'Width of die in WF_UNITS              ', 'Missing' : 0.0   },
                'WF_UNITS' : {'#' :  6, 'Type' : 'U*1', 'Ref' : None, 'Value' : None, 'Text' : 'Units for wafer and die dimensions    ', 'Missing' : 0     }, # 0=?/1=Inch/2=cm/3=mm/4=mils
                'WF_FLAT'  : {'#' :  7, 'Type' : 'C*1', 'Ref' : None, 'Value' : None, 'Text' : 'Orientation of wafer flat (U/D/L/R)   ', 'Missing' : ' '   },
                'CENTER_X' : {'#' :  8, 'Type' : 'I*2', 'Ref' : None, 'Value' : None, 'Text' : 'X coordinate of center die on wafer   ', 'Missing' : -32768},
                'CENTER_Y' : {'#' :  9, 'Type' : 'I*2', 'Ref' : None, 'Value' : None, 'Text' : 'Y coordinate of center die on wafer   ', 'Missing' : -32768},
                'POS_X'    : {'#' : 10, 'Type' : 'C*1', 'Ref' : None, 'Value' : None, 'Text' : 'Positive X direction of wafer (L/R)   ', 'Missing' : ' '   },
                'POS_Y'    : {'#' : 11, 'Type' : 'C*1', 'Ref' : None, 'Value' : None, 'Text' : 'Positive Y direction of wafer (U/D)   ', 'Missing' : ' '   }
            }
        else:
            raise STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
        self._default_init(endian, record)

    def to_atdf(self):

        sequence = {}
        header = ''
        body = ''
        
        header = self.id + ':'

#        The order of fields is different in STDF and ATDF for WCR record
#        
#        STDF page 40| ATDF page 36 
#        
#        3  WAFR_SIZ        
#        4  DIE_HT    
#        5  DIE_WID
#        6  WF_UNITS       
#        7  WF_FLAT     = 7 WF_FLAT
#        8  CENTER_X    
#        9  CENTER_Y    
#        10 POS_X       = 10 POS_X
#        11 POS_Y       = 11 POS_Y
#                          3 WAFR_SIZ
#                          4 DIE_HT
#                          5 DIE_WID
#                          6 WF_UNITS
#                          8 CENTER_X
#                          9 CENTER_Y
                      
#       7 WF_FLAT
        body += self.gen_atdf(7)
#       10 POS_X
        body += self.gen_atdf(10)
#       11 POS_Y
        body += self.gen_atdf(11)
#       3 WAFR_SIZ
        body += self.gen_atdf(3)
#       4 DIE_HT
        body += self.gen_atdf(4)
#       5 DIE_WID
        body += self.gen_atdf(5)
#       6 WF_UNITS
        body += self.gen_atdf(6)
#       8 CENTER_X
        body += self.gen_atdf(8)
#       9 CENTER_Y
        body += self.gen_atdf(9)
        
        body = body[:-1] 

        # assemble the record
        retval = header + body

        if self.local_debug: print("%s._to_atdf()\n   '%s'\n" % (self.id, retval))
        return retval


import sys
from . import STDR


class PRR(STDR):
    def __init__(self, version=None, endian=None, record=None):
        self.id = 'PRR'
        self.local_debug = False
        if version==None or version=='V4':
            self.version = 'V4'
            self.info = '''
Part Results Record
-------------------

Function:
    Contains the result information relating to each part tested by the test program. The
    PRR and the Part Information Record (PIR) bracket all the stored information
    pertaining to one tested part.

Frequency:
    * Obligatory
    * One per part tested.

Location:
    Anywhere in the data stream after the corresponding PIR and before the MRR.
    Sent after completion of testing each part.
'''
            self.fields = {
                'REC_LEN'  : {'#' :  0, 'Type' : 'U*2', 'Ref' : None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' :                                     None},
                'REC_TYP'  : {'#' :  1, 'Type' : 'U*1', 'Ref' : None, 'Value' :    5, 'Text' : 'Record type                           ', 'Missing' :                                     None},
                'REC_SUB'  : {'#' :  2, 'Type' : 'U*1', 'Ref' : None, 'Value' :   20, 'Text' : 'Record sub-type                       ', 'Missing' :                                     None},
                'HEAD_NUM' : {'#' :  3, 'Type' : 'U*1', 'Ref' : None, 'Value' : None, 'Text' : 'Test head number                      ', 'Missing' :                                        1},
                'SITE_NUM' : {'#' :  4, 'Type' : 'U*1', 'Ref' : None, 'Value' : None, 'Text' : 'Test site number                      ', 'Missing' :                                        1},
# changed by seimit : bits 5 - 7: Reserved for future use — must be 0, page 44 from STDF v4
#               'PART_FLG' : {'#' :  5, 'Type' : 'B*1', 'Ref' : None, 'Value' : None, 'Text' : 'Part information flag                 ', 'Missing' : ['0', '0', '0', '1', '0', '0', '0', '1']},
                'PART_FLG' : {'#' :  5, 'Type' : 'B*1', 'Ref' : None, 'Value' : None, 'Text' : 'Part information flag                 ', 'Missing' : ['0', '0', '0', '1', '0', '0', '0', '0']},
                'NUM_TEST' : {'#' :  6, 'Type' : 'U*2', 'Ref' : None, 'Value' : None, 'Text' : 'Number of tests executed              ', 'Missing' :                                        0},
                'HARD_BIN' : {'#' :  7, 'Type' : 'U*2', 'Ref' : None, 'Value' : None, 'Text' : 'Hardware bin number                   ', 'Missing' :                                        0},
                'SOFT_BIN' : {'#' :  8, 'Type' : 'U*2', 'Ref' : None, 'Value' : None, 'Text' : 'Software bin number                   ', 'Missing' :                                    65535},
                'X_COORD'  : {'#' :  9, 'Type' : 'I*2', 'Ref' : None, 'Value' : None, 'Text' : '(Wafer) X coordinate                  ', 'Missing' :                                   -32768},
                'Y_COORD'  : {'#' : 10, 'Type' : 'I*2', 'Ref' : None, 'Value' : None, 'Text' : '(Wafer) Y coordinate                  ', 'Missing' :                                   -32768},
                'TEST_T'   : {'#' : 11, 'Type' : 'U*4', 'Ref' : None, 'Value' : None, 'Text' : 'Elapsed test time in milliseconds     ', 'Missing' :                                        0},
                'PART_ID'  : {'#' : 12, 'Type' : 'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Part identification                   ', 'Missing' :                                       ''},
                'PART_TXT' : {'#' : 13, 'Type' : 'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Part description text                 ', 'Missing' :                                       ''},
                'PART_FIX' : {'#' : 14, 'Type' : 'B*n', 'Ref' : None, 'Value' : None, 'Text' : 'Part repair information               ', 'Missing' :                                       []}
            }

        else:
            raise STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
        self._default_init(endian, record)

    def to_atdf(self):

        sequence = {}
        header = ''
        body = ''
        
        header = self.id + ':'

#        The order of fields is different in STDF and ATDF for PRR record
#        
#        STDF page 43| ATDF page 39
#        
#         3 HEAD_NUM    =  3 HEAD_NUM    
#         4 SITE_NUM    =  4 SITE_NUM  
#         5 PART_FLG    
#         6 NUM_TEST        
#         7 HARD_BIN       
#         8 SOFT_BIN       
#         9 X_COORD       
#        10 Y_COORD       
#        11 TEST_T       
#        12 PART_ID     = 12 PART_ID
#                          6 NUM_TEST
#                          5 PART_FLG bits 3 & 4
#                          7 HARD_BIN
#                          8 SOFT_BIN
#                          9 X_COORD
#                         10 Y_COORD
#                          5 PART_FLG bit 0 or 1
#                         5 PART_FLG bit 2
#                         11 TEST_T
#        13 PART_TXT    = 13 PART_TXT
#        14 PART_FIX    = 14 PART_FIX
    
#       3 HEAD_NUM 
        body += self.gen_atdf(3)
#       4 SITE_NUM 
        body += self.gen_atdf(4)
#       12 PART_ID 
        body += self.gen_atdf(12)
#       6 NUM_TEST 
        body += self.gen_atdf(6)
        
#                             5 PART_FLG bits 3 & 4
#           bit 4: 0 = Pass/fail flag (bit 3) is valid
        v = self.get_fields(5)[3] 
        if v != None and v[4] == '0':
            # 0 = Part passed
            if self.get_fields(5)[3][3] == '0':
                body += 'P|'
            # 1 = Part failed
            elif self.get_fields(5)[3][3] == '1':
                body += 'F|'
#           bit 4: 1 = Device completed testing with no pass/fail indication (i.e., bit 3 is invalid)
        elif v != None and v[4] == '1':
            buff += ' |'
#                              7 HARD_BIN
        body += self.gen_atdf(7)
#                              8 SOFT_BIN
        body += self.gen_atdf(8)
#                              9 X_COORD
        body += self.gen_atdf(9)
#                             10 Y_COORD
        body += self.gen_atdf(10)
#                              5 PART_FLG bit 0 or 1
#           Note: Either Bit 0 or Bit 1 can be set, but not both. 
        v = self.get_fields(5)[3] 

        if v != None and v[0] == '1' and v[1] == '1':
            raise STDFError("export to atdf error: PRR PART_FLG bits 0 and 1 are both set, which is not allowed" )
        elif v != None and v[0] == '1':
            body += "I|"
        elif v != None and v[1] == '1':
            body += "C|"
#                              5 PART_FLG bit 2
        if v != None and v[2] == '0':
            body += "|"
        elif v != None and v[2] == '1':
            body += "Y|"
#                             11 TEST_T
        body += self.gen_atdf(11)
#                             13 PART_TXT
        body += self.gen_atdf(13)
#                             14 PART_FIX
        value = self.get_fields(14)[3]
        if value != None:
            # converting bits into HEX values
            byte_list = []
            byte_value = 0
            bit = 0
            for i in range(len(value)):
                el = value[i]
                byte_value |= int(el) << bit
                bit += 1
                if bit == 8:
                    byte_list.append(byte_value)
                    byte_value = 0
                    bit = 0
            # ToDo : to make real n*bits
    #            if bit != 0:
            final_value = ''
            for elem in byte_list:
                final_value += hex(elem)[2:]
            body += final_value.upper()

        # assemble the record
        retval = header + body

        if self.local_debug: print("%s._to_atdf()\n   '%s'\n" % (self.id, retval))
        return retval
    


import sys
from . import STDR

class MPR(STDR):
    def __init__(self, version=None, endian=None, record=None):
        self.id = 'MPR'
        self.local_debug = False
        if version==None or version == 'V4':
            self.version ='V4'
            self.info = '''
Multiple-Result Parametric Record
---------------------------------

Function:
    Contains the results of a single execution of a parametric test in the test program
    where that test returns multiple values. The first occurrence of this record also
    establishes the default values for all semi-static information about the test, such as
    limits, units, and scaling. The MPR is related to the Test Synopsis Record (TSR) by test
    number, head number, and site number.

Frequency:
    * Obligatory, one per multiple-result parametric test execution on each head/site

Location:
    Anywhere in the data stream after the corresponding Part Information Record (PIR)
    and before the corresponding Part Result Record (PRR).
'''
            self.fields = {
                'REC_LEN'  : {'#' :  0, 'Type' : 'U*2',  'Ref' :       None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' :                                     None},
                'REC_TYP'  : {'#' :  1, 'Type' : 'U*1',  'Ref' :       None, 'Value' :   15, 'Text' : 'Record type                           ', 'Missing' :                                     None},
                'REC_SUB'  : {'#' :  2, 'Type' : 'U*1',  'Ref' :       None, 'Value' :   15, 'Text' : 'Record sub-type                       ', 'Missing' :                                     None},
                'TEST_NUM' : {'#' :  3, 'Type' : 'U*4',  'Ref' :       None, 'Value' : None, 'Text' : 'Test number                           ', 'Missing' :                                     None},
                'HEAD_NUM' : {'#' :  4, 'Type' : 'U*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Test head number                      ', 'Missing' :                                        1},
                'SITE_NUM' : {'#' :  5, 'Type' : 'U*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Test site number                      ', 'Missing' :                                        1},
                'TEST_FLG' : {'#' :  6, 'Type' : 'B*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Test flags (fail, alarm, etc.)        ', 'Missing' :                                  ['0']*8},
                'PARM_FLG' : {'#' :  7, 'Type' : 'B*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Parametric test flags (drift, etc.)   ', 'Missing' : ['1', '1', '0', '0', '0', '0', '0', '0']}, # 0xC0
                'RTN_ICNT' : {'#' :  8, 'Type' : 'U*2',  'Ref' :       None, 'Value' : None, 'Text' : 'Count (j) of PMR indexes              ', 'Missing' :                                        0},
                'RSLT_CNT' : {'#' :  9, 'Type' : 'U*2',  'Ref' :       None, 'Value' : None, 'Text' : 'Count (k) of returned results         ', 'Missing' :                                        0},
                'RTN_STAT' : {'#' : 10, 'Type' : 'xN*1', 'Ref' : 'RTN_ICNT', 'Value' : None, 'Text' : 'Array of j returned states            ', 'Missing' :                                       []}, # RTN_ICNT = 0
                'RTN_RSLT' : {'#' : 11, 'Type' : 'xR*4', 'Ref' : 'RSLT_CNT', 'Value' : None, 'Text' : 'Array of k returned results           ', 'Missing' :                                       []}, # RSLT_CNT = 0
                'TEST_TXT' : {'#' : 12, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Descriptive text or label             ', 'Missing' :                                       ''},
                'ALARM_ID' : {'#' : 13, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Name of alarm                         ', 'Missing' :                                       ''},
                'OPT_FLAG' : {'#' : 14, 'Type' : 'B*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Optional data flag See note           ', 'Missing' : ['0', '0', '0', '0', '0', '0', '1', '0']}, # 0x02
                'RES_SCAL' : {'#' : 15, 'Type' : 'I*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Test result scaling exponent          ', 'Missing' :                                        0}, # OPT_FLAG bit 0 = 1
                'LLM_SCAL' : {'#' : 16, 'Type' : 'I*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Test low limit scaling exponent       ', 'Missing' :                                        0}, # OPT_FLAG bit 4 or 6 = 1
                'HLM_SCAL' : {'#' : 17, 'Type' : 'I*1',  'Ref' :       None, 'Value' : None, 'Text' : 'Test high limit scaling exponent      ', 'Missing' :                                        0}, # OPT_FLAG bit 5 or 7 = 1
                'LO_LIMIT' : {'#' : 18, 'Type' : 'R*4',  'Ref' :       None, 'Value' : None, 'Text' : 'Test low limit value                  ', 'Missing' :                                      0.0}, # OPT_FLAG bit 4 or 6 = 1
                'HI_LIMIT' : {'#' : 19, 'Type' : 'R*4',  'Ref' :       None, 'Value' : None, 'Text' : 'Test high limit value                 ', 'Missing' :                                      0.0}, # OPT_FLAG bit 5 or 7 = 1
                'START_IN' : {'#' : 20, 'Type' : 'R*4',  'Ref' :       None, 'Value' : None, 'Text' : 'Starting input value [condition]      ', 'Missing' :                                      0.0}, # OPT_FLAG bit 1 = 1
                'INCR_IN'  : {'#' : 21, 'Type' : 'R*4',  'Ref' :       None, 'Value' : None, 'Text' : 'Increment of input condition          ', 'Missing' :                                       -1}, # OPT_FLAG bit 1 = 1
                'RTN_INDX' : {'#' : 22, 'Type' : 'xU*2', 'Ref' : 'RTN_ICNT', 'Value' : None, 'Text' : 'Array of j PMR indexes                ', 'Missing' :                                       []}, # RTN_ICNT = 0
                'UNITS'    : {'#' : 23, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Units of returned results             ', 'Missing' :                                       ''},
                'UNITS_IN' : {'#' : 24, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'Input condition units                 ', 'Missing' :                                       ''},
                'C_RESFMT' : {'#' : 25, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'ANSI C result format string           ', 'Missing' :                                       ''},
                'C_LLMFMT' : {'#' : 26, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'ANSI C low limit format string        ', 'Missing' :                                       ''},
                'C_HLMFMT' : {'#' : 27, 'Type' : 'C*n',  'Ref' :       None, 'Value' : None, 'Text' : 'ANSI C high limit format string       ', 'Missing' :                                       ''},
                'LO_SPEC'  : {'#' : 28, 'Type' : 'R*4',  'Ref' :       None, 'Value' : None, 'Text' : 'Low specification limit value         ', 'Missing' :                                      0.0}, # OPT_FLAG bit 2 = 1
                'HI_SPEC'  : {'#' : 29, 'Type' : 'R*4',  'Ref' :       None, 'Value' : None, 'Text' : 'High specification limit value        ', 'Missing' :                                      0.0}  # OPT_FLAG bit 3 = 1
            }
        else:
            raise STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
        self._default_init(endian, record)

    def to_atdf(self):

        sequence = {}
        header = ''
        body = ''
        
        header = self.id + ':'
        
        #    The order of fields is different in STDF and ATDF for FTR record
            
        #    STDF page 53| ATDF page 47
            
        #     3 TEST_NUM    =  3 TEST_NUM   
        #     4 HEAD_NUM    =  4 HEAD_NUM    
        #     5 SITE_NUM    =  5 SITE_NUM  
        #     6 TEST_FLG    
        #     7 PARM_FLG            
        #     8 RTN_ICNT    -> missing
        
        #     9 RSLT_CNT    -> missing    
        #    10 RTN_STAT    = 10 RTN_STAT  
        #    11 RTN_RSLT    = 11 RTN_RSLT
        #                -> 6 TEST_FLG bits 6 & 7 and 7 PARM_FLG bit 5
        #                -> 6 TEST_FLG bits 0, 2, 3, 4 & 5 and 7 PARM_FLG bits 0, 1, 2, 3 & 4
        #    12 TEST_TXT    = 12 TEST_TXT 
        #    13 ALARM_ID    = 13 ALARM_ID 
        #                -> 7 PARM_FLG bits 6 & 7 
        #    14 OPT_FLAG 
        #    15 RES_SCAL
        #    16 LLM_SCAL
        #    17 HLM_SCAL
        #    18 LO_LIMIT
        #    19 HI_LIMIT
        #    20 START_IN
        #    21 INCR_IN
        #    22 RTN_INDX
        #    23 UNITS       = 23 UNITS 
        #    24 UNITS_IN
        #    25 C_RESFMT
        #    26 C_LLMFMT
        #    27 C_HLMFMT
        #    28 LO_SPEC
        #    29 HI_SPEC
        #                = 18 LO_LIMIT
        #                = 19 HI_LIMIT
        #                = 20 START_IN
        #                = 21 INCR_IN
        #                = 24 UNITS_IN
        #                = 22 RTN_INDX
        #                = 25 C_RESFMT
        #                = 26 C_LLMFMT
        #                = 27 C_HLMFMT
        #                = 28 LO_SPEC
        #                = 29 HI_SPEC
        #                = 15 RES_SCAL
        #                = 16 LLM_SCAL
        #                = 17 HLM_SCAL

#       3 TEST_NUM 
        body += self.gen_atdf(3)
#       4 HEAD_NUM 
        body += self.gen_atdf(4)
#       5 SITE_NUM  
        body += self.gen_atdf(5)
#       10 RTN_STAT 
        body += self.gen_atdf(10)
#       11 RTN_RSLT   MUST be float!!!
        value = self.get_fields(11)[3]
        if value != None:
            for elem in value:
                leading = round(elem,0)
                len_lead = len(str(leading))
                result = round(elem, 9 - len_lead)
                body += "%s," % result
            body = body[:-1] 
        body += "|"        
#       6 TEST_FLG bits 6 & 7
#           bit 6: Pass/fail flag (bit 7) is valid
        v = self.get_fields(6)[3]
        
        if v != None and v[6] == '0':
#           bit 7:
#           0 = Part passed
            if self.get_fields(6)[3][7] == '0':
                body += 'P'
#           1 = Part failed
            elif self.get_fields(6)[3][7] == '1':
                body += 'F'
#       7 PARM_FLG bit 5
        v = self.get_fields(7)[3] 
        if v != None and v[5] == '0':
            body += '|'
        elif v != None and self.get_fields(7)[3][5] == '1':
            body += 'A|'            
#       6 TEST_FLG bits 0, 2, 3, 4 & 5
#       bit 0:
#       0 = No alarm
        v = self.get_fields(6)[3] 
        if v != None and v[0] == '0':
            body += ''
#       1 = Alarm detected during testing
        elif v != None and v[0] == '1':
            body += 'A'
#       bit 2:
#       0= Test result is reliable
        if v != None and v[2] == '0':
            body += ''
#       1 = Test result is unreliable
        elif v != None and v[2] == '1':
            body += 'U'
#       bit 3:
#       0 = No timeout
        if v != None and v[3] == '0':
            body += ''
#       1 = Timeout occurred
        elif v != None and v[3] == '1':
            body += 'T'
#       bit 4:
#       0 = Test was executed
        if v != None and v[4] == '0':
            body += ''
#       1 = Test not executed
        elif v != None and v[4] == '1':
            body += 'N'
#       bit 5:
#       0 = No abort
        if v != None and v[5] == '0':
            body += ''
#       1= Test aborted
        elif v != None and v[5] == '1':
            body += 'X'
#       7 PARM_FLG bits 0, 1, 2, 3 & 4
#       bit 0:
#       0 = No scale error
        v = self.get_fields(7)[3] 
        if v != None and v[0] == '0':
            body += ''
#       1 = Scale error
        elif v != None and v[0] == '1':
            body += 'S'
#       bit 1:
#       0 = No drift error
        if v != None and v[1] == '0':
            body += ''
#       1 = Drift error (unstable measurement)
        elif v != None and v[1] == '1':
            body += 'D'
#       bit 2:
#       0 = No oscillation
        if v != None and v[2] == '0':
            body += ''
#       1 = Oscillation detected
        elif v != None and v[2] == '1':
            body += 'O'
#       bit 3:
#       0 = Measured value not high
        if v != None and v[3] == '0':
            body += ''
#       1 = Measured value higher than high test limit
        elif v != None and v[3] == '1':
            body += 'H'
#       bit 4:
#       0 = Measured value not low
        if v != None and v[4] == '0':
            body += '|'
#       1 = Measured value lower than low test limit
        elif v != None and v[4] == '1':
            body += 'L|'
            
#       12 TEST_TXT
        body += self.gen_atdf(12)
#       13 ALARM_ID
        body += self.gen_atdf(13)

#       7 PARM_FLG bits 6 & 7
#       bit 6:
#       0 = If result = low limit, then result is “fail.”
        v = self.get_fields(7)[3] 
        if v != None and v[6] == '0':
            body += ''
#       1 = If result = low limit, then result is “pass.”
        elif v != None and v[6] == '1':
            body += 'L'
#       bit 7:
#       0 = If result = high limit, then result is “fail.”
        if v != None and v[7] == '0':
            body += '|'
#       1 = If result = high limit, then result is “pass.”
        elif v != None and v[7] == '1':
            body += 'H|'

#       23 UNITS         
        body += self.gen_atdf(23)
#       18 LO_LIMIT
        body += self.gen_atdf(18)
#       19 HI_LIMIT
        body += self.gen_atdf(19)
#       20 START_IN
        body += self.gen_atdf(20)
#       21 INCR_IN
        body += self.gen_atdf(21)
#       24 UNITS_IN
        body += self.gen_atdf(24)
#       22 RTN_INDX
        body += self.gen_atdf(22)
#       25 C_RESFMT
        body += self.gen_atdf(25)
#       26 C_LLMFMT
        body += self.gen_atdf(26)
#       27 C_HLMFMT
        body += self.gen_atdf(27)
#       28 LO_SPEC
        body += self.gen_atdf(28)
#       29 HI_SPEC
        body += self.gen_atdf(29)
#       15 RES_SCAL
        body += self.gen_atdf(15)
#       16 LLM_SCAL
        body += self.gen_atdf(16)
#       17 HLM_SCAL
        body += self.gen_atdf(17)
        body = body[:-1] 

        # assemble the record
        retval = header + body

        if self.local_debug: print("%s._to_atdf()\n   '%s'\n" % (self.id, retval))
        return retval


import sys
from . import STDR

class PMR(STDR):
    def __init__(self, version=None, endian=None, record = None):
        self.id = 'PMR'
        self.local_debug = False
        if version==None or version=='V4':
            self.version = 'V4'
            self.info =     '''
Pin Map Record
--------------

Function:
    Provides indexing of tester channel names, and maps them to physical and logical pin
    names. Each PMR defines the information for a single channel/pin combination.

Frequency:
    * Optional
    * One per channel/pin combination used in the test program.
    * Reuse of a PMR index number is not permitted.

Location:
    After the initial "FAR-(ATRs)-MIR-(RDR)-(SDRs)" sequence and before the first PGR, PLR,
    FTR, or MPR that uses this record's PMR_INDX value.
'''
            self.fields = {
                'REC_LEN'  : {'#' : 0, 'Type' :  'U*2', 'Ref' : None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' : None},
                'REC_TYP'  : {'#' : 1, 'Type' :  'U*1', 'Ref' : None, 'Value' :    1, 'Text' : 'Record type                           ', 'Missing' : None},
                'REC_SUB'  : {'#' : 2, 'Type' :  'U*1', 'Ref' : None, 'Value' :   60, 'Text' : 'Record sub-type                       ', 'Missing' : None},
                'PMR_INDX' : {'#' : 3, 'Type' :  'U*2', 'Ref' : None, 'Value' : None, 'Text' : 'Unique index associated with pin      ', 'Missing' :    0},
                'CHAN_TYP' : {'#' : 4, 'Type' :  'U*2', 'Ref' : None, 'Value' : None, 'Text' : 'Channel type                          ', 'Missing' :    0},
                'CHAN_NAM' : {'#' : 5, 'Type' :  'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Channel name                          ', 'Missing' :   ''},
                'PHY_NAM'  : {'#' : 6, 'Type' :  'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Physical name of pin                  ', 'Missing' :   ''},
                'LOG_NAM'  : {'#' : 7, 'Type' :  'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Logical name of pin                   ', 'Missing' :   ''},
                'HEAD_NUM' : {'#' : 8, 'Type' :  'U*1', 'Ref' : None, 'Value' : None, 'Text' : 'Head number associated with channel   ', 'Missing' :    1},
                'SITE_NUM' : {'#' : 9, 'Type' :  'U*1', 'Ref' : None, 'Value' : None, 'Text' : 'Site number associated with channel   ', 'Missing' :    1}
            }
        else:
            raise STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
        self._default_init(endian, record)


import sys
from . import STDR

class SBR(STDR):
    def __init__(self, version=None, endian=None, record=None):
        self.id = 'SBR'
        self.local_debug = False
        if version==None or version=='V4':
            self.version = 'V4'
            self.info = '''
Software Bin Record
-------------------

Function:
    Stores a count of the parts associated with a particular logical bin after testing. This
    bin count can be for a single test site (when parallel testing) or a total for all test sites.
    The STDF specification also supports a Hardware Bin Record (HBR) for actual physical
    binning. A part is "physically" placed in a hardware bin after testing. A part can be
    "logically" associated with a software bin during or after testing.

Frequency:
    * Obligatory
    * One per software bin for each head/site combination
    * One per software bin for all head/site combinations together ('HEAD_NUM' = 255)
    * May be included to name unused bins.

Location:
    Anywhere in the data stream after the initial "FAR-(ATRs)-MIR-(RDR)-(SDRs)" sequence and before the MRR.
    When data is being recorded in real time, this record usually appears near the end of the data stream.
'''
            self.fields = {
                'REC_LEN'  : {'#' :  0, 'Type' :  'U*2', 'Ref' : None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' : None},
                'REC_TYP'  : {'#' :  1, 'Type' :  'U*1', 'Ref' : None, 'Value' :    1, 'Text' : 'Record type                           ', 'Missing' : None},
                'REC_SUB'  : {'#' :  2, 'Type' :  'U*1', 'Ref' : None, 'Value' :   50, 'Text' : 'Record sub-type                       ', 'Missing' : None},
                'HEAD_NUM' : {'#' :  3, 'Type' :  'U*1', 'Ref' : None, 'Value' : None, 'Text' : 'Test head number (255 = summary)      ', 'Missing' : 1   },
                'SITE_NUM' : {'#' :  4, 'Type' :  'U*1', 'Ref' : None, 'Value' : None, 'Text' : 'Test site number                      ', 'Missing' : 1   },
                'SBIN_NUM' : {'#' :  5, 'Type' :  'U*2', 'Ref' : None, 'Value' : None, 'Text' : 'Software bin number                   ', 'Missing' : 0   },
                'SBIN_CNT' : {'#' :  6, 'Type' :  'U*4', 'Ref' : None, 'Value' : None, 'Text' : 'Number of parts in bin                ', 'Missing' : 0   },
                'SBIN_PF'  : {'#' :  7, 'Type' :  'C*1', 'Ref' : None, 'Value' : None, 'Text' : 'Pass/fail indication (P/F)            ', 'Missing' : ' ' },
                'SBIN_NAM' : {'#' :  8, 'Type' :  'C*n', 'Ref' : None, 'Value' : None, 'Text' : 'Name of software bin                  ', 'Missing' : ''  }
            }

        else:
            raise STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
        self._default_init(endian, record)


import sys
from . import STDR

class PLR(STDR):
    def __init__(self, version=None, endian=None, record = None):
        self.id = 'PLR'
        self.local_debug = False
        if version==None or version=='V4':
            self.version = 'V4'
            self.info = '''
Pin List Record
---------------

Function:
    Defines the current display radix and operating mode for a pin or pin group.

Frequency:
    * Optional
    * One or more whenever the usage of a pin or pin group changes in the test program.

Location:
    After all the PMRs and PGRs whose PMR index values and pin group index values are
    listed in the GRP_INDX array of this record; and before the first FTR that references pins
    or pin groups whose modes are defined in this record.
'''
            self.fields = {
                'REC_LEN'  : {'#' :  0, 'Type' :  'U*2', 'Ref' :       None, 'Value' : None, 'Text' : 'Bytes of data following header        ', 'Missing' : None},
                'REC_TYP'  : {'#' :  1, 'Type' :  'U*1', 'Ref' :       None, 'Value' :    1, 'Text' : 'Record type                           ', 'Missing' : None},
                'REC_SUB'  : {'#' :  2, 'Type' :  'U*1', 'Ref' :       None, 'Value' :   63, 'Text' : 'Record sub-type                       ', 'Missing' : None},
                'GRP_CNT'  : {'#' :  3, 'Type' :  'U*2', 'Ref' :       None, 'Value' : None, 'Text' : 'Count (k) of pins or pin groups       ', 'Missing' :    0},
                'GRP_INDX' : {'#' :  4, 'Type' : 'xU*2', 'Ref' :  'GRP_CNT', 'Value' : None, 'Text' : 'Array of pin or pin group indexes     ', 'Missing' :   []},
                'GRP_MODE' : {'#' :  5, 'Type' : 'xU*2', 'Ref' :  'GRP_CNT', 'Value' : None, 'Text' : 'Operating mode of pin group           ', 'Missing' :   []},
                'GRP_RADX' : {'#' :  6, 'Type' : 'xU*1', 'Ref' :  'GRP_CNT', 'Value' : None, 'Text' : 'Display radix of pin group            ', 'Missing' :   []},
                'PGM_CHAR' : {'#' :  7, 'Type' : 'xC*n', 'Ref' :  'GRP_CNT', 'Value' : None, 'Text' : 'Program state encoding characters     ', 'Missing' :   []},
                'RTN_CHAR' : {'#' :  8, 'Type' : 'xC*n', 'Ref' :  'GRP_CNT', 'Value' : None, 'Text' : 'Return state encoding characters      ', 'Missing' :   []},
                'PGM_CHAL' : {'#' :  9, 'Type' : 'xC*n', 'Ref' :  'GRP_CNT', 'Value' : None, 'Text' : 'Program state encoding characters     ', 'Missing' :   []},
                'RTN_CHAL' : {'#' : 10, 'Type' : 'xC*n', 'Ref' :  'GRP_CNT', 'Value' : None, 'Text' : 'Return state encoding characters      ', 'Missing' :   []}
            }
        else:
            raise STDFError("%s object creation error: unsupported version '%s'" % (self.id, version))
        self._default_init(endian, record)

    def to_atdf(self):

        sequence = {}
        header = ''
        body = ''
        
        header = self.id + ':'

#        The order of fields is different in STDF and ATDF for PLR record
#        
#        STDF page 32| ATDF page 27
#        
#          3 GRP_CNT    -> missing
#          4 GRP_INDX   = 4 GRP_INDX
#          5 GRP_MODE   = 5 GRP_MODE
#          6 GRP_RADX   = 6 GRP_RADX
#          7 PGM_CHAR
#          8 RTN_CHAR
#          9 PGM_CHAL
#         10 RTN_CHAL
#                       = 9 PGM_CHAL , 7 PGM_CHAR
#                       = 10 RTN_CHAL, 8 RTN_CHAR

#       4 GRP_INDX
        body += self.gen_atdf(4)
#       5 GRP_MODE
        body += self.gen_atdf(5)
#       6 GRP_RADX
        body += self.gen_atdf(6)

        pgm_right = self.get_fields(7)[3]
        pgm_left = self.get_fields(9)[3]
        
        if pgm_right == None and pgm_left == None:
            body += '|'
            pass
        else:
            if len(pgm_left) == 0:
#               7 PGM_CHAR
                body += self.gen_atdf(7)
            else:
                for right, left in zip(pgm_right, pgm_left):
                    body += right + "," + left +'/'
                body = body[:-1]
                body += '|'

        rtn_right = self.get_fields(8)[3]
        rtn_left = self.get_fields(10)[3]
        
        if rtn_right == None and rtn_left == None:
            body += '|'
            pass
        else:
            if len(rtn_left) == 0:
    #           8 RTN_CHAR
                body += self.gen_atdf(8)
            else:
                for right, left in zip(rtn_right, rtn_left):
                    body += right + "," + left +'/'
        body = body[:-1]

        # assemble the record
        retval = header + body

        if self.local_debug: print("%s._to_atdf()\n   '%s'\n" % (self.id, retval))
        return retval